use std::sync::Arc;

use flax::{entity_ids, BoxedSystem, Query, QueryBorrow, Schedule, System, World};
use vulkano::{
    command_buffer::{
        allocator::{CommandBufferAllocator, StandardCommandBufferAllocator},
        CommandBufferBeginInfo, CommandBufferLevel, CommandBufferUsage, RecordingCommandBuffer,
        RenderingAttachmentInfo, RenderingInfo,
    },
    image::view::ImageView,
    pipeline::graphics::viewport::Viewport,
    render_pass::{AttachmentLoadOp, AttachmentStoreOp},
    sync::GpuFuture,
};
use vulkano_util::{
    context::VulkanoContext, renderer::VulkanoWindowRenderer, window::VulkanoWindows,
};

use crate::core::{debug_gui::DebugGuiStack, module::RenderModule as ThreadLocalModule};

use self::{egui::Gui, test_pipeline::test_pipeline};

pub mod egui;
pub mod events;
mod test_pipeline;

pub struct RenderModule {
    schedule: Schedule,
    command_buffer_allocator: Arc<dyn CommandBufferAllocator>,
    viewport: Viewport,
}

impl RenderModule {
    pub fn new(
        vk_context: &mut VulkanoContext,
        _vk_windows: &mut VulkanoWindows,
        _schedule: &mut Schedule,
        _world: &mut World,
        _events: &mut crate::core::events::Events,
    ) -> Self {
        let schedule = Schedule::builder()
            .with_system(add_distance_system())
            .build();

        let command_buffer_allocator = Arc::new(StandardCommandBufferAllocator::new(
            vk_context.device().clone(),
            Default::default(),
        ));

        let viewport = Viewport {
            offset: [0.0, 0.0],
            extent: [0.0, 0.0],
            depth_range: 0.0..=1.0,
        };

        Self {
            schedule,
            command_buffer_allocator,
            viewport,
        }
    }
}

impl ThreadLocalModule for RenderModule {
    fn on_update(
        &mut self,
        gui_stack: &mut DebugGuiStack,
        vk_context: &mut VulkanoContext,
        vk_windows: &mut vulkano_util::window::VulkanoWindows,
        world: &mut World,
        _events: &mut crate::core::events::Events,
        _frame_time: std::time::Duration,
    ) -> anyhow::Result<()> {
        self.schedule.execute_seq(world).unwrap();

        let viewport = &mut self.viewport;

        for (window_id, renderer) in vk_windows.iter_mut() {
            let gui = gui_stack.get_mut(*window_id).unwrap();
            draw(
                self.command_buffer_allocator.clone(),
                viewport,
                vk_context,
                renderer,
                gui,
            );
        }
        Ok(())
    }
}

pub fn add_distance_system() -> BoxedSystem {
    let query = Query::new(entity_ids());

    System::builder()
        .with_query(query)
        .build(|mut query: QueryBorrow<'_, flax::EntityIds, _>| {
            for _id in &mut query {
                // println!("----------: {}", _id.index());
            }
        })
        .boxed()
}

fn draw(
    command_buffer_allocator: Arc<dyn CommandBufferAllocator>,
    viewport: &mut Viewport,
    context: &mut VulkanoContext,
    renderer: &mut VulkanoWindowRenderer,
    gui: &mut Gui,
) {
    let (vertex_buffer, pipeline) = test_pipeline(
        context.device().clone(),
        context.memory_allocator().clone(),
        renderer.swapchain_format(),
    );

    // Do not draw the frame when the screen size is zero. On Windows, this can
    // occur when minimizing the application.
    let image_extent: [u32; 2] = renderer.window().inner_size().into();

    if image_extent.contains(&0) {
        return;
    }

    // Begin rendering by acquiring the gpu future from the window renderer.
    let previous_frame_end = renderer
        .acquire(|swapchain_images| {
            // Whenever the window resizes we need to recreate everything dependent
            // on the window size. In this example that
            // includes the swapchain, the framebuffers
            // and the dynamic state viewport.
            window_size_dependent_setup(swapchain_images, viewport);
        })
        .unwrap();

    let mut builder = RecordingCommandBuffer::new(
        command_buffer_allocator.clone(),
        context.graphics_queue().queue_family_index(),
        CommandBufferLevel::Primary,
        CommandBufferBeginInfo {
            usage: CommandBufferUsage::OneTimeSubmit,
            ..Default::default()
        },
    )
    .unwrap();

    builder
        // Before we can draw, we have to *enter a render pass*. We specify which
        // attachments we are going to use for rendering here, which needs to match
        // what was previously specified when creating the pipeline.
        .begin_rendering(RenderingInfo {
            // As before, we specify one color attachment, but now we specify the image
            // view to use as well as how it should be used.
            color_attachments: vec![Some(RenderingAttachmentInfo {
                // `Clear` means that we ask the GPU to clear the content of this
                // attachment at the start of rendering.
                load_op: AttachmentLoadOp::Clear,
                // `Store` means that we ask the GPU to store the rendered output in
                // the attachment image. We could also ask it to discard the result.
                store_op: AttachmentStoreOp::Store,
                // The value to clear the attachment with. Here we clear it with a blue
                // color.
                //
                // Only attachments that have `AttachmentLoadOp::Clear` are provided
                // with clear values, any others should use `None` as the clear value.
                clear_value: Some([0.0, 0.0, 1.0, 1.0].into()),
                ..RenderingAttachmentInfo::image_view(
                    // We specify image view corresponding to the currently acquired
                    // swapchain image, to use for this attachment.
                    // attachment_image_views[image_index as usize].clone(),
                    renderer.swapchain_image_view().clone(),
                )
            })],
            ..Default::default()
        })
        .unwrap()
        // We are now inside the first subpass of the render pass.
        //
        // TODO: Document state setting and how it affects subsequent draw commands.
        .set_viewport(0, [viewport.clone()].into_iter().collect())
        .unwrap()
        .bind_pipeline_graphics(pipeline.clone())
        .unwrap()
        .bind_vertex_buffers(0, vertex_buffer.clone())
        .unwrap();

    unsafe {
        builder
            // We add a draw command.
            .draw(vertex_buffer.len() as u32, 1, 0, 0)
            .unwrap();
    }

    builder
        // We leave the render pass.
        .end_rendering()
        .unwrap();

    // Finish recording the command buffer by calling `end`.
    let command_buffer = builder.end().unwrap();

    draw_gui(gui);

    let before_future = previous_frame_end
        .then_execute(context.graphics_queue().clone(), command_buffer)
        .unwrap()
        .boxed();

    let after_future = gui
        .draw_on_image(before_future, renderer.swapchain_image_view())
        .boxed();

    // The color output is now expected to contain our triangle. But in order to
    // show it on the screen, we have to *present* the image by calling
    // `present` on the window renderer.
    //
    // This function does not actually present the image immediately. Instead it
    // submits a present command at the end of the queue. This means that it will
    // only be presented once the GPU has finished executing the command buffer
    // that draws the triangle.
    renderer.present(after_future, true);
}

fn draw_gui(gui: &mut Gui) {
    let mut code = CODE.to_owned();
    gui.immediate_ui(|gui| {
        let ctx = gui.context();
        egui::egui::Window::new("Colors")
            .vscroll(true)
            .show(&ctx, |ui| {
                ui.vertical_centered(|ui| {
                    ui.add(egui::egui::widgets::Label::new("Hi there!"));
                    sized_text(ui, "Rich Text", 32.0);
                });
                ui.separator();
                ui.columns(2, |columns| {
                    egui::egui::ScrollArea::vertical().id_source("source").show(
                        &mut columns[0],
                        |ui| {
                            ui.add(
                                egui::egui::TextEdit::multiline(&mut code)
                                    .font(egui::egui::TextStyle::Monospace),
                            );
                        },
                    );
                    egui::egui::ScrollArea::vertical()
                        .id_source("rendered")
                        .show(&mut columns[1], |ui| {
                            ui.add(egui::egui::widgets::Label::new("Good day!"));
                        });
                });
            });
    });
}

fn sized_text(ui: &mut egui::egui::Ui, text: impl Into<String>, size: f32) {
    ui.label(
        egui::egui::RichText::new(text)
            .size(size)
            .family(::egui::FontFamily::Monospace),
    );
}

const CODE: &str = r"
# Some markup
```
let mut gui = Gui::new(&event_loop, renderer.surface(), None, renderer.queue(), SampleCount::Sample1);
```
";

fn window_size_dependent_setup(
    image_views: &[Arc<ImageView>],
    viewport: &mut Viewport,
) -> Vec<Arc<ImageView>> {
    let extent = image_views[0].image().extent();
    viewport.extent = [extent[0] as f32, extent[1] as f32];

    image_views
        .iter()
        .map(|image_view| {
            let image = image_view.image().clone();
            ImageView::new_default(image).unwrap()
        })
        .collect::<Vec<_>>()
}
